package com.abi.qms.platform.service.impl;

import com.abi.qms.platform.dao.enums.QrTypeEnum;
import com.abi.qms.platform.dao.mapper.ReportServiceMapper;
import com.abi.qms.platform.dao.vo.result.ListBrandCodeReportVO;
import com.abi.qms.platform.dao.vo.result.ListFactoryCodeReportVO;
import com.abi.qms.platform.dao.vo.result.ListMaterialCodeReportVO;
import com.abi.qms.platform.dao.vo.result.ListPackageCodeVo;
import com.abi.qms.platform.dto.req.ReportReq;
import com.abi.qms.platform.dto.res.ListBrandCodeReportRes;
import com.abi.qms.platform.dto.res.ListFactoryCodeReportRes;
import com.abi.qms.platform.dto.res.ListMaterialCodeRankingReportRes;
import com.abi.qms.platform.dto.res.PackageCodeTypeRes;
import com.abi.qms.platform.service.ReportService;
import com.abi.task.common.utils.PojoConverterUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;
import org.springframework.util.ObjectUtils;

import java.text.DecimalFormat;
import java.time.DayOfWeek;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalAdjusters;
import java.util.*;
import java.math.BigDecimal;
import java.util.stream.Collectors;

/**
 * @author ludashi
 * @date 2021年06月02日 15:39
 */
@Service
public class ReportServiceImpl implements ReportService {

    @Autowired
    private ReportServiceMapper reportServiceMapper;

    /**
     * 品牌使用码量占比（603版 已废弃）
     *
     * @param reportReq
     * @return com.abi.qms.platform.dto.res.ListBrandCodeReportRes
     * @author ludashi
     * @date 2021/6/2 17:50
     */
    /*@Override
    public ListBrandCodeReportRes queryBrandCodeProportion(ReportReq reportReq) {
        //参数设置初始化
        setQueryTime(reportReq);
        //查询 品牌数量
        List<ListBrandCodeReportVO> listReportVOS = reportServiceMapper.queryBrandCodeProportion(reportReq);
        if(!CollectionUtils.isEmpty(listReportVOS)){
            //求总和
            Long collect = listReportVOS.stream().collect(Collectors.summingLong(ListBrandCodeReportVO::getQrNumber));
            listReportVOS.forEach(listReportVO->{
                Double val =listReportVO.getQrNumber().doubleValue() / collect.doubleValue()* 100;
                listReportVO.setPercent(new DecimalFormat("#0.00").format(val + 0.000001d)+"%");
            });
        }
        // 封装出参
        ListBrandCodeReportRes res = new ListBrandCodeReportRes();
        List<ListBrandCodeReportRes.BrandCodeReportBean> reportRes = PojoConverterUtils.copyList(listReportVOS, ListBrandCodeReportRes.BrandCodeReportBean.class);
        res.setReportBrandCodeResList(reportRes);
        return res;
    }*/

    /**
     * 物料使用码量排名
     * @param reportReq
     * @return
     */
    @Override
    public ListMaterialCodeRankingReportRes queryMaterialCodeRanking(ReportReq reportReq) {
        List<ListMaterialCodeReportVO> listMaterialCodeReportVOS = reportServiceMapper.queryMaterialCodeRanking(reportReq);
        // 封装出参
        ListMaterialCodeRankingReportRes res = new ListMaterialCodeRankingReportRes();
        List<ListMaterialCodeRankingReportRes.MaterialCodeReportResBean> materialCodeReportResBeans = PojoConverterUtils.copyList(listMaterialCodeReportVOS, ListMaterialCodeRankingReportRes.MaterialCodeReportResBean.class);
        res.setReportMaterialCodeRankingResList(materialCodeReportResBeans);
        return res;
    }


    /**
     * 包材厂使用码量，下载率
     * @param reportReq
     * @return
     */
    @Override
    public ListFactoryCodeReportRes queryFactoryCodeTotalReport(ReportReq reportReq) {
        //参数设置初始化
        setQueryTime(reportReq);
        //包材厂生成包数量
        List<ListFactoryCodeReportVO> listFactoryCodeTotalReportVOS = reportServiceMapper.queryFactoryCodeTotal(reportReq,0);
        //包材厂包下载量
        List<ListFactoryCodeReportVO> listFactoryCodeDownloadTotalReportVOS = reportServiceMapper.queryFactoryCodeTotal(reportReq,1);
        Map<String, Long> factoryCodeDownloadTotalMaps = listFactoryCodeDownloadTotalReportVOS.stream().collect(Collectors.toMap(ListFactoryCodeReportVO::getFactoryName, ListFactoryCodeReportVO::getQrNumber));
        if(!CollectionUtils.isEmpty(listFactoryCodeTotalReportVOS)) {
            listFactoryCodeTotalReportVOS.forEach(listReportVO -> {
                Double val = 0d;
                if (null != factoryCodeDownloadTotalMaps.get(listReportVO.getFactoryName())) {
                    val = factoryCodeDownloadTotalMaps.get(listReportVO.getFactoryName()).doubleValue() / listReportVO.getQrNumber().doubleValue() * 100;
                }
                listReportVO.setDownloadNumber(new DecimalFormat("#0.00").format(val + 0.000001d) + "%");
            });
        }
        // 封装出参
        ListFactoryCodeReportRes res = new ListFactoryCodeReportRes();
        List<ListFactoryCodeReportRes.FactoryCodeReportResBean> reportFactoryRes = PojoConverterUtils.copyList(listFactoryCodeTotalReportVOS, ListFactoryCodeReportRes.FactoryCodeReportResBean.class);
        res.setFactoryCodeResReportList(reportFactoryRes);
        return res;
    }

    /**
     * 设置时间区间值
     * @param reportReq
     */
    private void setQueryTime(ReportReq reportReq){
        if(!ObjectUtils.isEmpty(reportReq) &&
                null == reportReq.getEndTime() &&
                null == reportReq.getBeginTime() &&
                !ObjectUtils.isEmpty(reportReq.getValue())){
            HashMap<String, LocalDateTime> timestamp = getTimestamp(reportReq.getValue());
            reportReq.setBeginTime( timestamp.get("startTime"));
            reportReq.setEndTime( timestamp.get("endTime"));
        }
    }

    /**
     * 根据类型获取时间范围
     * @param timeType (0-当天，1-本周，2-本月)
     * @return key 开始时间：startTime  结束时间：endTime
     */
    private HashMap<String, LocalDateTime> getTimestamp(Integer timeType) {
        HashMap<String, LocalDateTime> hashMap = new HashMap<String, LocalDateTime>();
        Calendar calendar = Calendar.getInstance();

        //当天
        if (timeType == 0){
            hashMap.put("startTime", longTurnData(getTimestampByOffsetDay(0)));
            hashMap.put("endTime", longTurnData(getTimestampByOffsetDay(1)));
        }
        //本周
        if (timeType == 1){
            hashMap.put(
                    "startTime",
                    longTurnData(getTimestampByOffsetDay(0 - calendar.get(Calendar.DAY_OF_WEEK) + 2)) );
            hashMap.put(
                    "endTime",
                    longTurnData( getTimestampByOffsetDay(calendar
                            .getMaximum(Calendar.DAY_OF_WEEK)
                            - calendar.get(Calendar.DAY_OF_WEEK) + 2)));
        }

        //本月
        if (timeType == 2){
            hashMap.put(
                    "startTime",
                    longTurnData(getTimestampByOffsetDay(0 - calendar.get(Calendar.DAY_OF_MONTH) + 1)));
            hashMap.put(
                    "endTime",
                    longTurnData( getTimestampByOffsetDay(calendar
                            .getMaximum(Calendar.DAY_OF_MONTH)
                            - calendar.get(Calendar.DAY_OF_MONTH))));
        }
        return hashMap;
    }

    /**
     * 时间格式转化
     * @author ludashi
     * @date 2021/6/3 13:37
     * @param l 时间毫秒值
     * @return java.time.LocalDateTime
     */
    private LocalDateTime longTurnData(Long l) {
        Instant instant = Instant.ofEpochMilli(l);
        ZoneId zone = ZoneId.systemDefault();
        LocalDateTime localDateTime = LocalDateTime.ofInstant(instant, zone);
        return localDateTime;
    }

    /**
     * 时间定义
     * @author ludashi
     * @date 2021/6/3 13:12
     * @param day
     * @return long
     */
    public long getTimestampByOffsetDay(int day){
        Calendar calendar = Calendar.getInstance();
        calendar.add(Calendar.DAY_OF_MONTH, day);
        calendar.set(Calendar.HOUR_OF_DAY, 0);
        calendar.set(Calendar.SECOND, 0);
        calendar.set(Calendar.MINUTE, 0);
        calendar.set(Calendar.MILLISECOND, 0);
        return calendar.getTimeInMillis();
    }

    /**
     * 码包概况
     * @author liguosong
     * @date 2021/6/3
     * @param reportReq
     * @return
     */
    @Override
    public PackageCodeTypeRes queryPackageCodeGeneral(ReportReq reportReq) {
        Map<String,LocalDateTime> map=new HashMap<String,LocalDateTime>();
        //判断是否有参数值
        if(null!= reportReq.getValue()){
           map= FillParam(reportReq,map);
        }
        PackageCodeTypeRes res = new PackageCodeTypeRes();
        //查询同期生成码
        getGenerateCodeNum(reportReq.getBeginTime(),reportReq.getEndTime(),1,res,"1");
        //查询同期下载码
        getGenerateCodeNum(reportReq.getBeginTime(),reportReq.getEndTime(),2,res,"1");
        //查询同期激活码
        getActivateCodeNum(reportReq.getBeginTime(),reportReq.getEndTime(),res,"1");
        //查询往期
        if(!map.isEmpty()){
            //查询往期生成码
            getGenerateCodeNum(map.get("beginDate"),map.get("endDate"),1,res,"2");
            //查询往期下载码
            getGenerateCodeNum(map.get("beginDate"),map.get("endDate"),2,res,"2");
            //查询往期激活码
            getActivateCodeNum(map.get("beginDate"),map.get("endDate"),res,"2");
        }
        //返回参数
        return res;
    }

    /**
     * 封装入参
     * @param reportReq
     * @param map
     */
    private Map FillParam(ReportReq reportReq,Map<String,LocalDateTime> map) {
        setQueryTime(reportReq);
        map=getBeforeWeekTime(reportReq.getValue().toString());//1周 2月
        return map;
    }

    /**
     * 获取生成码和下载码情况
     * @param beginDate 开始时间
     * @param endDate 结束时间
     * @param type 判断是生成码还是下载码 1生成 2激活
     * @param res 返回值
     * @param value 判断是否往期
     */
    private void getGenerateCodeNum(LocalDateTime beginDate,LocalDateTime endDate,Integer type,PackageCodeTypeRes res,String value) {
        ListPackageCodeVo entityBase = reportServiceMapper.queryPackageCodeNum(beginDate,endDate,type);
        if(!ObjectUtils.isEmpty(entityBase)){
            //判断是否往期
            if("1".equals(value)){
                //判断是否生成码使用情况/或者下载码使用情况
                if("1".equals(type.toString())){
                    res.setCreateBoxNumber(entityBase.getQrBoxNumber());
                    res.setCreateCapNumber(entityBase.getQrCapNumber());
                }else {
                    res.setLoadBoxNumber(entityBase.getQrBoxNumber());
                    res.setLoadCapNumber(entityBase.getQrCapNumber());
                }
            }else{
                //判断是否生成码使用情况/或者下载码使用情况
                if("1".equals(type.toString())){
                    long present=res.getCreateBoxNumber()+res.getCreateCapNumber();
                    long before=entityBase.getQrBoxNumber()+entityBase.getQrCapNumber();
                    res.setCreatePercent(percentBigDecimal(present, before));
                }else{
                    long present=res.getLoadBoxNumber()+res.getLoadCapNumber();
                    long before=entityBase.getQrBoxNumber()+entityBase.getQrCapNumber();
                    res.setLoadPercent(percentBigDecimal(present, before));
                }
            }
        }

    }

    /**
     * 获取激活码情况
     * @param beginDate 开始时间
     * @param endDate 结束时间
     * @param res 出参
     * @param value 判断同期/往期
     */
    private void getActivateCodeNum(LocalDateTime beginDate,LocalDateTime endDate,PackageCodeTypeRes res,String value) {
        ListPackageCodeVo activateEntity = reportServiceMapper.queryActivateCodeNum(beginDate,endDate);
        //判断查询返回是否为空
        if(!ObjectUtils.isEmpty(activateEntity)){
            //判断同期还是往期
            if("1".equals(value)){
                res.setActivateBoxNumber(activateEntity.getQrBoxNumber());
                res.setActivateCapNumber(activateEntity.getQrCapNumber());
            }else{
                long present=res.getActivateBoxNumber()+res.getActivateCapNumber();
                long before=activateEntity.getQrBoxNumber()+activateEntity.getQrCapNumber();
                String result = percentBigDecimal(present, before);
                res.setActivatePercent(result);
            }
        }

    }

    /**
     * @param present 同期数
     * @param before 往期数
     * @return
     */
    public static String percentBigDecimal(long present, long before) {
        BigDecimal preNum = BigDecimal.valueOf(present);
        BigDecimal sufNum = BigDecimal.valueOf(before);
        double result = countDecimal(preNum, sufNum);
        if (result > 0) {
            return "+" + result + "%";
        }
        if (result < 0) {
            return "-" + Math.abs(result) + "%";
        }
        if (result == 0) {
            return "0%";
        }
        return null;
    }

    public static double countDecimal(BigDecimal preNum, BigDecimal sufNum) {
        boolean preBoolean = verifyNum(preNum);
        boolean sufBoolean = verifyNum(sufNum);
        //同时为true计算
        if (preBoolean && sufBoolean) {
            boolean b = verifyEqual(preNum, sufNum);
            if (b == false) {
                return realCountDecimal(preNum, sufNum);
            }
            if (b) {
                return 0;
            }
        }
        if (preBoolean == false && sufBoolean == false) {
            return 0;
        }
        if (sufBoolean == false) {
            return 100;
        }
        if (preBoolean == false) {
            return -100;
        }
        return 0;
    }

    //验证数字是否为零和null
    public static boolean verifyNum(BigDecimal num) {
        if (null != num && num.compareTo(BigDecimal.ZERO) != 0) {
            return true;
        }
        return false;
    }

    //验证两个数字是否相等
    public static boolean verifyEqual(BigDecimal preNum, BigDecimal sufNum) {
        int n = preNum.compareTo(sufNum);
        if (n == 0) {
            return true;
        }
        return false;
    }

    //真正计算
    public static double realCountDecimal(BigDecimal preNum, BigDecimal sufNum) {
        //(前面的数字-后面的数字)/后面的数字*100
        BigDecimal bigDecimal = (preNum.subtract(sufNum)).divide(sufNum, 2, BigDecimal.ROUND_HALF_UP).multiply(new BigDecimal("100")).setScale(2, BigDecimal.ROUND_HALF_UP);
        if (bigDecimal.compareTo(BigDecimal.ZERO) != 0) {
            return bigDecimal.doubleValue();
        }
        return 0;
    }

    /**
     * 获取往期时间（1-周，2-月）
     * @return
     */
    public Map<String,LocalDateTime> getBeforeWeekTime(String value){
        Map<String,LocalDateTime> map=new HashMap<String,LocalDateTime>();
        if("1".equals(value)){
            LocalDateTime now = LocalDateTime.now();
            LocalDateTime todayOfLastWeek = now.minusDays(7);
            map.put("beginDate",todayOfLastWeek.with(TemporalAdjusters.previous(DayOfWeek.SUNDAY)).plusDays(1));
            map.put("endDate",todayOfLastWeek.with(TemporalAdjusters.next(DayOfWeek.MONDAY)).minusDays(1));
        }else{
            DateTimeFormatter fmt = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss");
            LocalDateTime date2 = LocalDateTime.parse(fmt.format(LocalDateTime.now()), fmt);
            LocalDateTime localDateTime = date2.minusMonths(1);
            map.put("beginDate",localDateTime.with(TemporalAdjusters.firstDayOfMonth()));
            map.put("endDate",localDateTime.with(TemporalAdjusters.lastDayOfMonth()));
        }
        return map;
    }

}